library hakkit.hakk.builder;
import 'dart:io' show Platform, File, FileSystemException;

part 'src/blocks.dart';
part 'src/instructions.dart';

var _hakkContext = new StringBuffer();
final String CR = Platform.isWindows ? '\r\n' : '\r';

List<String> _blockNames = [];
List<String> _instrBlockNames = [];
List<String> _memBlockNames = [];
List<String> _codeBlockNames = [];
List<String> _staticInducerNames = [];
List<String> _thermalInducerNames = [];
List<String> _heatSinkNames = [];
List<String> _staticDischargerNames = [];
List<String> _rebooterNames = [];
List<String> _subHakkPortNames = [];
List<String> _hakkNames = [];

bool _blocksDeclared = false;
bool _codeDeclared = false;
bool _inSHPContext = false;
bool _inBlocksContext = false;
bool _inCodeContext = false;
bool _inCodeBlockContext = false;

/// A Hakk builder function that returns a hakkML source file in YAML format.
///
/// @param hakkName The name of the hakk.
/// @param body A function that encapsulates the other sections of the hakk.
/// @returns A string representing the generated hakkML.  Suitable for upload
/// and play at http://hakk.it
/// @throws String A string error detailing the reason for the failure.
String hakk(String hakkName, String body()){
  _init();
  _checkNotInCodeContext();
  _checkNotInBlockContext();
  _checkNotInCodeBlockContext();
  _checkNotInSHPContext();
  _hakkNames.add(hakkName);
  _hakkContext.write('name: "$hakkName"$CR');
  body();

  if (!_blocksDeclared){
    _error('No blocks were declared for this hakk');
  }

  if (!_codeDeclared){
    _error('No code was declared for this hakk.');
  }

  if(_instrBlockNames.any((name) => !_codeBlockNames.contains(name))){
    _error('The entry block and any declared instruction blocks, must also have a corresponding code block defined for them.');
  }

  return _hakkContext.toString();
}


/**
 * A helper function that allows using a strongly typed
 * representation of Core memory registers (a quite common thing in hakks).
 *
 * ## Usage
 * Instead of doing this in your declarations:
 *
 *     // adds the values in core.1 and core.2, stores in core.0
 *     I.add('core.0', 'core.1', 'core.2');
 *
 * you can use this class to do this:
 *
 *     I.add(core(0), core(1), core(2));
 *
 * and core will return the appropriate strings for you.
 */
String core(int register) => 'core.$register';


/// Saves the contents of a [hakkML] string to a given [pathAndNameNoExtension].
///
/// The method will append the ".hakk" extension to your file so there is no
/// need to provide that.
///
/// @param pathAndNameNoExtension The path and filename to save as.
/// @param hakkML The hakkML source generated by the [hakk] builder.
void saveToFile(String pathAndNameNoExtension, String hakkML){
  final f = new File('${pathAndNameNoExtension}.hakk');
  try{
    f.writeAsStringSync(hakkML);
  } on FileSystemException catch (e){
    print('The save operation failed: $e');
  }
}


_init(){
  _hakkContext = new StringBuffer();

  _blockNames.clear();
  _instrBlockNames.clear();
  _memBlockNames.clear();;
  _codeBlockNames.clear();
  _staticInducerNames.clear();
  _thermalInducerNames.clear();
  _heatSinkNames.clear();
  _staticDischargerNames.clear();
  _rebooterNames.clear();
  _subHakkPortNames.clear();
  _hakkNames.clear();

  _blocksDeclared = false;
  _codeDeclared = false;
  _inSHPContext = false;
  _inBlocksContext = false;
  _inCodeContext = false;
  _inCodeBlockContext = false;
  B._firstShpBlock = true;
}

void _checkNotInSHPContext(){
  if (_inSHPContext){
    _error('This element cannot be declared in a SubHakkPort (SHP).');
  }
}

void _checkInSHPContext(){
  if (!_inSHPContext){
    _error('This element can only be declared in a SubHakkPort (SHP)');
  }
}

void _checkNotInCodeContext(){
  if(_inCodeContext){
    _error('This element cannot be declared in the code section.');
  }
}

void _checkInBlockContext(){
  if (!_inBlocksContext){
    _error('This element can only be declared in the block section.');
  }
}

void _checkInCodeBlockContext(){
  if(!_inCodeBlockContext){
    _error('This element can only be declared in a code block.');
  }
}

void _checkInCodeContext(){
  if (!_inCodeContext){
    _error('This element can only be declared in the code section');
  }
}

void _checkNotInBlockContext(){
  if (_inBlocksContext){
    _error('This element cannot be declared in the blocks section.');
  }
}

void _checkNotInCodeBlockContext(){
  if (_inCodeBlockContext){
    _error('This element cannot be declared in an instruction block.');
  }
}

void _checkBlockExists(String name){
  if (_instrBlockNames.contains(name)){
    _error('There is already a block declared with this name: $name');
  }
}

void _writeBlock(String bName, String name, int x, int y){
  _blockNames.add(name);
  _hakkContext.write('  - $bName: {name: "$name", loc : [$x, $y]}$CR');
}

void _error(String message, {ctx: ''}){
  throw 'Hakk Builder Error: $message$CR${_hakkContext.toString()}$CR$ctx';
}