part of hakkit.hakk.builder;

void codeBlock(String name, body()){
  if (_inCodeContext == false){
    _error('A code block must be declared inside the code section.');
  }
  _checkNotInBlockContext();
  _checkNotInCodeBlockContext();
  _checkNotInSHPContext();

  _inCodeBlockContext = true;
  _codeBlockNames.add(name);
  _hakkContext.write('  $name:$CR');
  body();
  _inCodeBlockContext = false;
}

void code(void body()){
  _checkNotInCodeContext();
  _checkNotInBlockContext();
  _checkNotInCodeBlockContext();
  _checkNotInSHPContext();
  if(!_blocksDeclared){
    _error('The blocks section should be declared first before declaring code.');
  }
  if(_codeDeclared){
    _error('A hakkML file can only declare a single section which defines code blocks.');
  }
  _inCodeContext = true;
  _hakkContext.write('code:$CR');
  _codeDeclared = true;
  body();
  _inCodeContext = false;
}

/**
 * hakkML Instructions.
 */
class I
{
  /// Defines an addition instruction.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference of the first operand.
  /// @param val2 The literal number or memory reference of the second operand.
  static void add(String store, val1, val2){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - add: [$store, $val1, $val2]$CR');
  }

  /// Defines a subtraction instruction.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference of the first operand.
  /// @param val2 The literal number or memory reference of the second operand.
  static void sub(String store, val1, val2){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - sub: [$store, $val1, $val2]$CR');
  }

  /// Defines an multiplication instruction.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference of the first operand.
  /// @param val2 The literal number or memory reference of the second operand.
  static void mul(String store, val1, val2){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - mul: [$store, $val1, $val2]$CR');
  }

  /// Defines an division instruction.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference of the first operand.
  /// @param val2 The literal number or memory reference of the second operand.
  static void div(String store, val1, val2){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    if (val2 is num && val2 == 0){
      _error('Attempty to divide by 0.', ctx: 'second operand is $val2');
    }
    _hakkContext.write('    - div: [$store, $val1, $val2]$CR');
  }

  /// Defines a remark in the source code of the Hakk.
  /// This instruction is discarded at runtime.
  ///
  /// @param remark The remark.
  static void rem(String remark){
    _checkErrors();
    _hakkContext.write('    - rem: "$remark"$CR');
  }

  /// Causes Hakk execution to wait for a given amount of ticks
  /// before continuing.
  ///
  /// @param amountOfTicks The literal (or memory reference to
  /// literal) amount of ticks in the VM to wait.
  static void wait(amountOfTicks){
    _checkErrors();
    _checkIntOrReference(amountOfTicks);
    _hakkContext.write('    - wait: $amountOfTicks$CR');
  }

  /// Computes a vector from the position of the Core block of your Hakk to
  /// the given x & y coordinates in the playfield.
  ///
  /// The vector x/y results are pushed onto the stack, y first.
  ///
  /// @param x The literal number or memory reference of the x coordinate.
  /// @param y The literal number or memory reference of the y coordinate.
  static void tvec(x, y){
    _checkErrors();
    _checkIntOrReference(x);
    _checkIntOrReference(y);
    _hakkContext.write('    - tvec: [$x, $y]$CR');
  }

  /// Instructs the VM to turn your Hakk in a given direction.
  ///
  /// @param store The memory reference location to store the result of the
  /// turn (1 succeeded, 0 failed).
  /// @param direction The literal integer or memory reference indicating the
  /// direction to turn(1 for left, 0 for right).
  static void turn(store, direction){
    _checkErrors();
    _checkStore(store);
    _checkIntOrReference(direction);
    if (direction is int && (direction != 0 && direction != 1)){
      _error('If specified as a literal, the direction must be a 0(right) or 1(left).');
    }
    _hakkContext.write('    - turn: [$store, $direction]$CR');
  }

  /// Instructs the VM to move your Hakk forwards one unit on the playfield.
  ///
  /// @param store The memory reference location to store the result of the
  /// move (1 succeeded, 0 failed).
  static void movef(store){
    _checkErrors();
    _checkStore(store);
    _hakkContext.write('    - movef: $store$CR');
  }

  /// Instructs the VM to move your Hakk backwards one unit on the playfield.
  ///
  /// @param store The memory reference location to store the result of the
  /// move (1 succeeded, 0 failed).
  static void moveb(store){
    _checkErrors();
    _checkStore(store);
    _hakkContext.write('    - moveb: $store$CR');
  }

  /// Using a Static Inducer (si) block, introduces a static charge to a
  /// playfield cell at the given x/y coordinates for a period of time.  Hakks
  /// crossing into this cell will receive a computational performance
  /// penalty, for the duration of the effect.
  ///
  /// @params staticInducerName The name of the si what will deploy the static
  /// charge.  This block must already be declared in the blocks section of
  /// your Hakk.
  /// @param x The literal number or memory reference of the x coordinate.
  /// @param y The literal number or memory reference of the y coordinate.
  static void stat(String staticInducerName, x, y){
    _checkErrors();
    if (!_staticInducerNames.contains(staticInducerName)){
      _error('The static inducer name was not found declared in the blocks section.', ctx: staticInducerName);
    }
    _checkIntOrReference(x);
    _checkIntOrReference(y);
    _hakkContext.write('    - stat: [$staticInducerName, $x, $y]$CR');
  }

  /// Using a Thermal Inducer (ti) block, introduces heat to a
  /// playfield cell at the given x/y coordinates for a period of time.  Hakks
  /// crossing into this cell will heat up the longer they are exposed to the
  /// effects of the heat generated, until they become disabled.
  ///
  /// @params heatInducerName The name of the ti what will deploy the heat.
  /// This block must already be declared in the blocks section of your Hakk.
  ///
  /// @param x The literal number or memory reference of the x coordinate.
  /// @param y The literal number or memory reference of the y coordinate.
  static void heat(String thermalInducerName, x, y){
    _checkErrors();
    if (!_thermalInducerNames.contains(thermalInducerName)){
      _error('The thermal inducer name was not found declared in the blocks section.', ctx: thermalInducerName);
    }
    _checkIntOrReference(x);
    _checkIntOrReference(y);
    _hakkContext.write('    - heat: [$thermalInducerName, $x, $y]$CR');
  }


  /// Using a rebooter (rebt) block, this instruction can revive a disabled
  /// block.
  ///
  /// @param rebooterName The name of the rebt block that will perform the
  /// revive operation.  This block must already be declared in the blocks
  /// section of your Hakk.
  /// @param hakkName The name of the Hakk or Sub-Hakk that holds the target
  /// block.
  /// @param blockNameOrIndex  The name of the block to perform the cooling
  /// operation on, or the integer index to the block (in declared order) in
  /// the Hakk's block list.  Using the index is useful, for example, if you
  /// want to iterate through all blocks continuously as part of your
  /// defensive strategy.
  static void heal(String rebooterName, String hakkName, String blockNameOrIndex){
    _checkErrors();
    if (!rebooterName.contains('.') && !_rebooterNames.contains(rebooterName)){
      _error('The rebt name was not found declared in the blocks section.', ctx: rebooterName);
    }
    if (!hakkName.contains('.') && !_hakkNames.contains(hakkName)){
      _error('The hakk name you specified does not match the name you gave, or any of the sub-hakk names.', ctx: hakkName);
    }
    if (!blockNameOrIndex.contains('.') && blockNameOrIndex is String && !_blockNames.contains(blockNameOrIndex)){
      _error('It looks like you specified a literal block name target, but no name was found that matched it.', ctx: blockNameOrIndex);
    }

    if (rebooterName.contains('.')){
      _getMemoryReference(rebooterName);
    }

    if (hakkName.contains('.')){
      _getMemoryReference(hakkName);
    }

    if (blockNameOrIndex.contains('.')){
      _getMemoryReference(blockNameOrIndex);
    }

    _hakkContext.write('    - heal: [$rebooterName, $hakkName, $blockNameOrIndex]$CR');
  }


  /// Using a heat sink (hsnk) block, this instruction reduces the heat level
  /// on a given block. This instruction can address any hakk/sub-hakk that you
  /// have deployed in the playfield.
  ///
  /// @param heatSinkName The name of the hsnk block that will perform the
  /// cooling operation.  This block must already be declared in the blocks
  /// section of your Hakk.
  /// @param hakkName The name of the Hakk or Sub-Hakk that holds the target
  /// block.
  /// @param blockNameOrIndex  The name of the block to perform the cooling
  /// operation on, or the integer index to the block (in declared order) in
  /// the Hakk's block list.  Using the index is useful, for example, if you
  /// want to iterate through all blocks continuously as part of your
  /// defensive strategy.
  static void sink(String heatSinkName, String hakkName, String blockNameOrIndex){
    _checkErrors();
    if (!heatSinkName.contains('.') && !_heatSinkNames.contains(heatSinkName)){
      _error('The hsnk name was not found declared in the blocks section.', ctx: heatSinkName);
    }
    if (!hakkName.contains('.') && !_hakkNames.contains(hakkName)){
      _error('The hakk name you specified does not match the name you gave, or any of the sub-hakk names.', ctx: hakkName);
    }
    if (!blockNameOrIndex.contains('.') && blockNameOrIndex is String && !_blockNames.contains(blockNameOrIndex)){
      _error('It looks like you specified a literal block name target, but no name was found that matched it.', ctx: blockNameOrIndex);
    }

    if (heatSinkName.contains('.')){
      _getMemoryReference(heatSinkName);
    }

    if (hakkName.contains('.')){
      _getMemoryReference(hakkName);
    }

    if (blockNameOrIndex.contains('.')){
      _getMemoryReference(blockNameOrIndex);
    }

    _hakkContext.write('    - sink: [$heatSinkName, $hakkName, $blockNameOrIndex]$CR');
  }

  /// Using a static discharger (sdis) block, this instruction reduces the
  /// static level  on a given block. This instruction can address any
  /// hakk/sub-hakk that you have deployed in the playfield.
  ///
  /// @param heatSinkName The name of the hsnk block that will perform the
  /// cooling operation.  This block must already be declared in the blocks
  /// section of your Hakk.
  /// @param hakkName The name of the Hakk or Sub-Hakk that holds the target
  /// block.
  /// @param blockNameOrIndex  The name of the block to perform the cooling
  /// operation on, or the integer index to the block (in declared order) in
  /// the Hakk's block list.  Using the index is useful, for example, if you
  /// want to iterate through all blocks continuously as part of your
  /// defensive strategy.
  static void disc(String staticDischargerName, String hakkName, String blockNameOrIndex){
    _checkErrors();
    if (!staticDischargerName.contains('.') && !_staticDischargerNames.contains(staticDischargerName)){
      _error('The sdis name was not found declared in the blocks section.', ctx: staticDischargerName);
    }
    if (!hakkName.contains('.') && !_hakkNames.contains(hakkName)){
      _error('The hakk name you specified does not match the name you gave, or any of the sub-hakk names.', ctx: hakkName);
    }
    if (!blockNameOrIndex.contains('.') && blockNameOrIndex is String && !_blockNames.contains(blockNameOrIndex)){
      _error('It looks like you specified a literal block name target, but no name was found that matched it.', ctx: blockNameOrIndex);
    }

    if (staticDischargerName.contains('.')){
      _getMemoryReference(staticDischargerName);
    }

    if (hakkName.contains('.')){
      _getMemoryReference(hakkName);
    }

    if (blockNameOrIndex.contains('.')){
      _getMemoryReference(blockNameOrIndex);
    }

    _hakkContext.write('    - disc: [$staticDischargerName, $hakkName, $blockNameOrIndex]$CR');
  }


  /// Brpt (Block Report) pushes state information about a block onto
  /// the stack:
  ///
  /// pushes amount of static effect the block has
  /// pushes amount of heat buildup the block has
  /// pushes 1 or 0 indicating whether the block is disabled (1 = disabled)
  ///
  /// A stop error is issued if the hakk or block is not found, or some other
  /// problem occurs.
  ///
  /// @param hakkName The name of the Hakk or Sub-Hakk that holds the target
  /// block.
  /// @param blockNameOrIndex  The name of the block to generate the report for,
  /// or the integer index to the block (in declared order) in
  /// the Hakk's block list.  Using the index is useful, for example, if you
  /// want to iterate through all blocks continuously as part of your
  /// play strategy.
  static void brpt(String hakkName, String blockNameOrIndex){
    _checkErrors();
    if (!hakkName.contains('.') && !_hakkNames.contains(hakkName)){
      _error('The hakk name you specified does not match the name you gave, or any of the sub-hakk names.', ctx: hakkName);
    }
    if (!blockNameOrIndex.contains('.') && blockNameOrIndex is String && !_blockNames.contains(blockNameOrIndex)){
      _error('It looks like you specified a literal block name target, but no name was found that matched it.', ctx: blockNameOrIndex);
    }

    if (hakkName.contains('.')){
      _getMemoryReference(hakkName);
    }

    if (blockNameOrIndex.contains('.')){
      _getMemoryReference(blockNameOrIndex);
    }

    _hakkContext.write('    - brpt: [$hakkName, $blockNameOrIndex]$CR');
  }



  /// Performs a cosine operation on a given value and stores the result
  /// in the given memory reference.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference to the value.
  static void cos(String store, val){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val);

    _hakkContext.write('    - cos: [$store, $val]$CR');
  }

  /// Performs a sine operation on a given value and stores the result
  /// in the given memory reference.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference to the value.
  static void sin(String store, val){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val);

    _hakkContext.write('    - sin: [$store, $val]$CR');
  }

  /// Performs a sine operation on a given value and stores the result
  /// in the given memory reference.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val1 The literal number or memory reference to the value.
  static void pow(String store, val, pow){
    _checkErrors();
    _checkStore(store);
    _checkNumOrReference(val);
    _checkIntOrReference(pow);

    _hakkContext.write('    - pow: [$store, $val, $pow]$CR');
  }


  /// Launches a sub-hakk into the playfield.
  ///
  /// @param store The memory reference location to store the result of the
  /// launch (1 succeeded, 0 failed).
  /// @param subHakkPortName The name of the shp block to trigger the launch
  /// from.  This block must be previously defined in the blocks section
  /// of your Hakk.
  static void shpl(String store, String subHakkPortName){
    _checkErrors();
    _checkStore(store);
    if (!_subHakkPortNames.contains(subHakkPortName)){
      _error('A shp block was not found with the given name.', ctx: subHakkPortName);
    }
    _hakkContext.write('    - shpl: [$store, $subHakkPortName]$CR');
  }

  /// Sets a memory reference in [store] with the given value.
  /// This instruction is also useful for copying a value from one
  /// memory location to another since [val] can also be a memory
  /// reference.
  ///
  /// @param store The memory reference location to store the result.
  /// @param val The literal, or memory reference to the literal, that should
  /// be written to the store memory reference.
  static void set(String store, val){
    _checkErrors();
    _checkStore(store);
    if (val is String && val.contains('.')){
      _getMemoryReference(val);
    }
    _hakkContext.write('    - set: [$store, $val]$CR');
  }

  /// Scans a location on the playfield and returns a result indicating which
  /// type of block is located at that position.  If not block is located at
  /// the location, then -1 is returned.
  ///
  /// @param store The memory reference location to store the result.
  /// @param x The literal integer or memory reference to the value of the x
  /// coordinate.
  /// @param y The literal integer or memory reference to the value of the y
  /// coordinate.
  static void scan(String store, x, y){
    _checkErrors();
    _checkStore(store);
    _checkIntOrReference(x);
    _checkIntOrReference(y);

    _hakkContext.write('    - scan: [$store, $x, $y]$CR');
  }

  /// Calculates the straight line distance (rounded) from the Hakk core block
  /// location to the given x/y coordinates on the playfield.  If the coordinate
  /// values are invalid, then -1 will be stored.
  ///
  /// @param store The memory reference location to store the result.
  /// @param x The literal integer or memory reference to the value of the x
  /// coordinate.
  /// @param y The literal integer or memory reference to the value of the y
  /// coordinate.
  static void dist(String store, x, y){
    _checkErrors();
    _checkStore(store);
    _checkIntOrReference(x);
    _checkIntOrReference(y);

    _hakkContext.write('    - dist: [$store, $x, $y]$CR');
  }



  /// Generates a random integer between 0 (inclusive) and [val] (exclusive)
  /// and stores it in [store].
  ///
  /// @param store The memory reference location to store the result.
  /// @param val The literal integer or memory reference to the value of the
  /// high boundary number.
  static void rnd(String store, val){
    _checkErrors();
    _checkStore(store);
    _checkIntOrReference(val);

    _hakkContext.write('    - rnd: [$store, $val]$CR');
  }

  /// Returns control flow to the instruction after the most recent `call`
  /// instruction.
  static void ret(){
    _checkErrors();
    _hakkContext.write('    - ret:$CR');
  }

  /// Pushes a value onto the stack of the Hakk's core block.
  ///
  /// @param val The literal or memory reference to the value to push to the
  /// stack.
  static void push(val){
    _checkErrors();
    _checkNumOrReference(val);
    _hakkContext.write('    - push: $val$CR');
  }

  /// Pops a value from the stack of the Hakk's core block.
  ///
  /// @param store The memory reference where the popped value should be stored.
  static void pop(store){
    _checkErrors();
    _checkStore(store);
    _hakkContext.write('    - pop: $store$CR');
  }

  /// Stores the number of blocks in the Hakk.
  ///
  /// @param store The memory reference where the popped value should be stored.
  static void bamt(store){
    _checkErrors();
    _checkStore(store);
    _hakkContext.write('    - bamt: $store$CR');
  }

  /// Declares a label that can be used in control flow operations as a target
  /// point.
  ///
  /// @param name The name of the label.
  static void lbl(String name){
    _checkErrors();
    _hakkContext.write('    - lbl: $name$CR');
  }

  /// Changes control flow to the given [jumpToName] (can be a label or a
  /// instruction block name), if equality exists between [val1] and [val2].
  ///
  /// @param val1 The first value, or memory reference to the value, to compare.
  /// @param val2 The second value, or memory reference to the value, to compare.
  /// @param jumpToName The label or instruction block name to jump to if the
  /// condition is satisfied.
  static void je(val1, val2, String jumpToName){
    _checkErrors();
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - je: [$val1, $val2, $jumpToName]$CR');
  }

  /// Changes control flow to the given [jumpToName] (can be a label or a
  /// instruction block name), if [val1] is greater than [val2].
  ///
  /// @param val1 The first value, or memory reference to the value, to compare.
  /// @param val2 The second value, or memory reference to the value, to compare.
  /// @param jumpToName The label or instruction block name to jump to if the
  /// condition is satisfied.
  static void jg(val1, val2, String jumpToName){
    _checkErrors();
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - jg: [$val1, $val2, $jumpToName]$CR');
  }

  /// Changes control flow to the given [jumpToName] (can be a label or a
  /// instruction block name), if [val1] is greater than or equal to [val2].
  ///
  /// @param val1 The first value, or memory reference to the value, to compare.
  /// @param val2 The second value, or memory reference to the value, to compare.
  /// @param jumpToName The label or instruction block name to jump to if the
  /// condition is satisfied.
  static void jge(val1, val2, String jumpToName){
    _checkErrors();
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - jge: [$val1, $val2, $jumpToName]$CR');
  }

  /// Changes control flow to the given [jumpToName] (can be a label or a
  /// instruction block name), if [val1] is less than [val2].
  ///
  /// @param val1 The first value, or memory reference to the value, to compare.
  /// @param val2 The second value, or memory reference to the value, to compare.
  /// @param jumpToName The label or instruction block name to jump to if the
  /// condition is satisfied.
  static void jl(val1, val2, String jumpToName){
    _checkErrors();
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - jl: [$val1, $val2, $jumpToName]$CR');
  }

  /// Changes control flow to the given [jumpToName] (can be a label or a
  /// instruction block name), if [val1] is less than or equal to [val2].
  ///
  /// @param val1 The first value, or memory reference to the value, to compare.
  /// @param val2 The second value, or memory reference to the value, to compare.
  /// @param jumpToName The label or instruction block name to jump to if the
  /// condition is satisfied.
  static void jle(val1, val2, String jumpToName){
    _checkErrors();
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - jle: [$val1, $val2, $jumpToName]$CR');
  }

  /// Changes control flow to the given [jumpToName] (can be a label or a
  /// instruction block name), if [val1] is not equal to [val2].
  ///
  /// @param val1 The first value, or memory reference to the value, to compare.
  /// @param val2 The second value, or memory reference to the value, to compare.
  /// @param jumpToName The label or instruction block name to jump to if the
  /// condition is satisfied.
  static void jne(val1, val2, String jumpToName){
    _checkErrors();
    _checkNumOrReference(val1);
    _checkNumOrReference(val2);
    _hakkContext.write('    - jne: [$val1, $val2, $jumpToName]$CR');
  }

  /// Sends the literal or memory reference to the value to stdout.
  ///
  /// WARNING: Be careful using string literals as the val parameter.  If
  /// the literal contains a '.', then it will be interpreted as a memory
  /// reference.
  ///
  /// @param val A literal value or a memory reference to a value.
  static void out(val){
    _checkErrors();
    if (val is String && val.contains('.')){
      _getMemoryReference(val);
    }
    if (val is String && !val.contains('.')){
      _hakkContext.write('    - out: "$val"$CR');
    }else{
      _hakkContext.write('    - out: $val$CR');
    }
  }

  /// Does nothing. Burns one tick in the VM.  This is equivalent to:
  ///     wait: 1
  static void nop(){
    _checkErrors();
    _hakkContext.write('    - nop:$CR');
  }

  /// Pushes the normalized x/y vector of the Hakk this instruction is
  /// executing in onto the stack, y first. This vector can be interpreted as
  /// the orientation of the Hakk on the playfield.
  static void hvec(){
    _checkErrors();
    _hakkContext.write('    - hvec:$CR');
  }

  /// Pushes the x/y location of the Hakk to the stack, y first.  This location
  /// is the absolute position of the core block in the Hakk on the playfield.
  static void hloc(){
    _checkErrors();
    _hakkContext.write('    - hloc:$CR');
  }

  /// A high cost instruction that will locate core block of the nearest
  /// opponent on the playfield.  The location is then pushed to the stack,
  /// y first.
  static void find(){
    _checkErrors();
    _hakkContext.write('    - find:$CR');
  }

  /// Changes control flow to a given label or instruction block name.
  ///
  /// @param gotoName The name of the label or instruction block name to goto.
  static void goto(String gotoName){
    _checkErrors();
    _hakkContext.write('    - goto: $gotoName$CR');
  }

  /// Changes control flow to a given label or instruction block name.  A call
  /// differs from a goto in that it preserves the location of the next
  /// instruction after the call in the call stack.  This means that call
  /// instructions are expected to always have a corresponding `ret` instruction
  /// somewhere else in the program execution (they should balance).
  ///
  /// Nested call/ret's are supported by the VM.  Recursion is also supported.
  ///
  /// @param callName The name of the label or instruction block name to call.
  static void call(String callName){
    _checkErrors();
    _hakkContext.write('    - call: $callName$CR');
  }

  /// Immediately stops program execution of a Hakk with a given
  /// exit integer value.
  ///
  /// @param val A literal value or a memory reference to a value.
  static void exit(val){
    _checkErrors();
    _checkIntOrReference(val);
    _hakkContext.write('    - exit: $val$CR');
  }

  static void _checkIntOrReference(ref){
    if (ref is int) return;
    _getMemoryReference(ref);
  }

  static void _checkNumOrReference(ref){
    if (ref is num) return;
    _getMemoryReference(ref);
  }

  static void _checkStore(String store){
    final ref = _getMemoryReference(store);
  }

  static Map<String, int> _getMemoryReference(String ref){
    if (ref is! String){
      _error('Expected the memory reference to be a string.', ctx: ref);
    }

    final pieces = ref.split('.');
    if (pieces.length != 2){
      _error('Expect two components in reference (name and index), found ${pieces.length}');
    }
    final name = pieces[0];

    if (!_memBlockNames.contains(name)){
      _error("The first component of a memory reference didn't match any memory blocks you declared (or the core block)", ctx: ref);
    }

    int index;
    try{
      index = int.parse(pieces[1]);
      if (index < 0)
      {
        _error('Reference index cannot be less than 0', ctx: ref);
      }
    } on FormatException catch (e){
      _error('The second component in a memory reference must be an integer index value.', ctx: ref);
    }

    return {name : index};
  }

  static void _checkErrors(){
    _checkNotInBlockContext();
    _checkNotInSHPContext();
    _checkInCodeContext();
    _checkInCodeBlockContext();
  }
}